tool
extends Node2D


const CHAIN_TOLERANCE = 0.01
const CHAIN_MAX_ITER = 10

export (Array, NodePath) var bones_in_fabrik_chain;
var bone_nodes = [];
var bone_node_transforms = [];
export (bool) var auto_calculate_bone_lengths = true;
export (Array, float) var bones_in_fabrik_chain_lengths;

export (bool) var auto_calculate_bone_angles = true;
export (Array, float) var bones_in_fabrik_chain_angles;

export (NodePath) var target_path = null;
var _target = null;
var _target_transform_to_use = null;

var _fabrik_chain_origin = Vector3();
var _fabrik_chain_total_length = INF;

export(int) var chain_iterations = 0;
export(bool) var limit_chain_iterations = true;
export(bool) var reset_iterations_on_update = true;
export (Array, Vector2) var bone_magnet_positions;

export (bool) var apply_joint_constraints = true;
export (Array, Vector2) var bone_angle_constraints;


func _ready():
	
	if (target_path != null):
		_target = get_node(target_path);
	else:
		print ("Twisted 2D FABRIK ERROR: ", name, " does not have a target!");
	
	# Get the bone nodes
	while (bones_in_fabrik_chain_lengths.size() < bones_in_fabrik_chain.size()):
		bones_in_fabrik_chain_lengths.append(Vector2.ZERO);
	
	for i in range(0, bones_in_fabrik_chain.size()):
		var bone_node = get_node(bones_in_fabrik_chain[i])
		if (bone_node != null):
			if bone_node is Node2D:
				bone_nodes.append(bone_node)
				if (auto_calculate_bone_lengths == true):
					if ("bone_length" in bone_node):
						bones_in_fabrik_chain_lengths[i] = bone_node.bone_length;
					else:
						if (i < bones_in_fabrik_chain.size()-1):
							var next_bone_node = get_node(bones_in_fabrik_chain[i+1]);
							bones_in_fabrik_chain_lengths[i] = (
								bone_node.global_transform.origin -
								next_bone_node.global_transform.origin
							).length();
			else:
				print ("Twisted 2D Constrained FABRIK ERROR: ", name,
				" has a nodepath for a bone that points to non Node2D node");
			continue;
		else:
			print ("Twisted 2D Constrained FABRIK Warning: ", name,
				" has a nodepath for a bone that does not point to a valid node");
			continue;
	
	# Populate the bone node transforms
	# and calculate the bone angles.
	for i in range(0, bone_nodes.size()):
		if (i >= bone_node_transforms.size()):
			bone_node_transforms.append(bone_nodes[i].global_transform);
		else:
			bone_node_transforms[i] = bone_nodes[i].global_transform;
		
		if (i >= bones_in_fabrik_chain_angles.size()):
			bones_in_fabrik_chain_angles.append(0);
		if (auto_calculate_bone_angles == true):
			if ("bone_angle" in bone_nodes[i]):
				bones_in_fabrik_chain_angles[i] = (bone_nodes[i].bone_angle);
			elif (i < bone_nodes.size()-1):
				var tmp_pos = bone_nodes[i].to_local(bone_nodes[i+1].global_position).normalized();
				var bone_angle = rad2deg(atan2(tmp_pos.y, tmp_pos.x));
				bones_in_fabrik_chain_angles[i] = (bone_angle);
	
	# Set the total length of the bone chain, if it is not already set
	_fabrik_chain_total_length = 0
	for bone_length in bones_in_fabrik_chain_lengths:
		_fabrik_chain_total_length += bone_length


func evaluate():
	if reset_iterations_on_update == true:
		chain_iterations = 0
	solve_chain()


func solve_chain():
	if chain_iterations >= CHAIN_MAX_ITER and limit_chain_iterations == true:
		return
	else:
		chain_iterations = 0
	
	# Get updated positions
	# (Might be better in the while loop?)
	for i in range(0, bone_nodes.size()):
		bone_node_transforms[i] = bone_nodes[i].global_transform;
	
	_fabrik_chain_origin = bone_nodes[0].global_transform.origin
	
	# Get the direction of the final bone by using the next to last bone if
	# there is more than 2 bones. If there are only 2 bones, we use the target's
	# forward Z vector instead (not ideal, but it works fairly well)
	var dir
	if bone_node_transforms.size() > 2:
		var dir_rot = (
			bone_node_transforms[bone_node_transforms.size()-1].origin -
			_target.global_position
		)
		var dir_angle = dir_rot.angle();
		dir = Vector2(
			cos(dir_angle),
			sin(dir_angle)
		)
	else:
		dir = Vector2(
			cos(_target.rotation),
			sin(_target.rotation)
		);
	
	# I'm not totally sure what/how target_pos is being used here...
	var target_pos = _target.global_transform.origin;
	target_pos += (dir * bones_in_fabrik_chain_lengths[bones_in_fabrik_chain_lengths.size()-1]);
	_target_transform_to_use = _target.global_transform;
	
	# Get the difference between our end effector (the final bone in the chain) and the target
	var dif = (bone_node_transforms[bone_node_transforms.size()-1].origin - target_pos).length()
	
	# Check to see if the distance from the end effector to the target is within our error margin (CHAIN_TOLERANCE).
	# If it not, move the chain towards the target (going forwards, backwards, and then applying rotation)
	while dif > CHAIN_TOLERANCE:
		
		# Apply bone magnet positions
		# TODO: make this work with the first bone in the chain.
		# TODO: make this work when the target is close.
		if (bone_magnet_positions.size() <= bone_node_transforms.size() and bone_magnet_positions.size() > 0):
			for i in range(0, bone_magnet_positions.size()):
				if (i == 0):
					pass;
				elif (i >= bone_nodes.size()-1):
					_target_transform_to_use.origin += bone_magnet_positions[i];
				else:
					bone_node_transforms[i].origin += bone_magnet_positions[i];
		
		chain_backward()
		chain_forward()
		
		# Update the difference between our end effector (the final bone in the chain) and the target
		dif = (bone_node_transforms[bone_node_transforms.size()-1].origin - target_pos).length()
		
		# Add one to chain_iterations. If we have reached our max iterations, then break
		chain_iterations = chain_iterations + 1
		if chain_iterations >= CHAIN_MAX_ITER:
			break


# Backward reaching pass
# (Is mislabled and should be considered the FORWARD reaching pass?)
func chain_backward():
	# Get the direction of the final bone by using the next to last bone if there is
	# more than 2 bones. If there are only 2 bones, we use the target's direction
	# vector instead (not ideal, but it works fairly well)
	#
	# NOTE: Interestingly, changing which node you use for calculating
	# the direction changes the angles used by the solver. Right now,
	# I am using a chain-like direction, though I was previously using
	# an "arm-like" direction.
	# Maybe make this an option in the future?
	var dir
	if bone_node_transforms.size() > 2:
		var dir_rot = (
			bone_node_transforms[bone_node_transforms.size()-1].origin -
			_target_transform_to_use.origin
		)
		var dir_angle = dir_rot.angle();
		dir = Vector2(
			cos(dir_angle),
			sin(dir_angle)
		)
	else:
		dir = Vector2(
			cos(_target.rotation),
			sin(_target.rotation)
		);
	
	var end_effector_index = bone_node_transforms.size()-1;
	
	# Set the rotation of the end effector
	# TODO: account for constraints.
	bone_node_transforms[end_effector_index] = _transform2d_lookat(
		bone_node_transforms[end_effector_index],
		_target_transform_to_use.origin
	);
	
	# Set the position of the end effector (the final bone in the chain)
	# to the target position
	bone_node_transforms[bone_node_transforms.size()-1].origin = (
		_target_transform_to_use.origin +  
		(dir * bones_in_fabrik_chain_lengths[bone_node_transforms.size()-1])
	)
	
	# Set the joint transform to the node.
	set_bone_transform(end_effector_index, bone_node_transforms[end_effector_index]);
	
	# For all of the other bones, move them towards the target
	var i = bone_node_transforms.size() - 1
	while i >= 1:
		var prev_origin = bone_node_transforms[i].origin
		i -= 1
		var curr_origin = bone_node_transforms[i].origin
		
		# ---- Start applying joint constraints ----
		if (apply_joint_constraints == true):
			if (i < bone_angle_constraints.size()):
				
				var direction_vector = prev_origin.direction_to(curr_origin);
				var direction_length = (prev_origin - curr_origin).length();
				var direction_angle = atan2(direction_vector.y, direction_vector.x);
				
				var two_bones_back_angle = 0;
				two_bones_back_angle = _calculate_previous_bones_angle(i)
				
				direction_angle -= deg2rad(180);
				
				var constrained_angle = _assure_angle_positive(direction_angle);
				constrained_angle = _clamp_rotation(
					constrained_angle,
					_assure_angle_positive(deg2rad(bone_angle_constraints[i].x) + two_bones_back_angle),
					_assure_angle_positive(deg2rad(bone_angle_constraints[i].y) + two_bones_back_angle)
				);
				
				constrained_angle += deg2rad(180);
				
				if (Input.is_action_just_pressed("ui_accept") == true):
					print ("I is: ", i);
					print ("Backwards constrained angle: ", rad2deg(constrained_angle));
					print ("Backwards bone angle constraints: ", Vector2(
						(bone_angle_constraints[i].x) + rad2deg(two_bones_back_angle),
						(bone_angle_constraints[i].y) + rad2deg(two_bones_back_angle)
					))
					print ("Backwards 2 bones back angle: ", rad2deg(two_bones_back_angle));
					
					print ("----------");
				
				var constrained_vector = Vector2(cos(constrained_angle), sin(constrained_angle));
				
				curr_origin = prev_origin + (constrained_vector * direction_length);
				bone_node_transforms[i].origin = curr_origin;
		# ---- No longer applying joint constraints. ----
		
		var r = prev_origin - curr_origin
		var l = bones_in_fabrik_chain_lengths[i] / r.length();
		var bone_final_pos = prev_origin.linear_interpolate(curr_origin, l)
		# Apply the position and rotation, (potentially with constraints), to the bone node
		bone_node_transforms[i] = _transform2d_lookat(bone_node_transforms[i], prev_origin);
		bone_node_transforms[i].origin = bone_final_pos
		
		# Set the joint transform to the node.
		set_bone_transform(i, bone_node_transforms[i]);


# Forward reaching pass
func chain_forward():
	# Set root at initial position
	bone_node_transforms[0].origin = _fabrik_chain_origin
	
	# Set the joint transform to the node.
	set_bone_transform(0, bone_node_transforms[0]);
	
	# Go through every bone in the bone chain
	for i in range(bone_node_transforms.size() - 1):
		var curr_origin = bone_node_transforms[i].origin
		var next_origin = bone_node_transforms[i+1].origin
		
		# ---- Start applying joint constraints ----
		if (apply_joint_constraints == true):
			if (i < bone_angle_constraints.size()):
				
				var direction_vector = curr_origin.direction_to(next_origin);
				var direction_length = (curr_origin - next_origin).length();
				var direction_angle = atan2(direction_vector.y, direction_vector.x);
				
				var two_bones_back_angle = 0;
				two_bones_back_angle = _calculate_previous_bones_angle(i)
				
				var constrained_angle = _assure_angle_positive(direction_angle);
				constrained_angle = _clamp_rotation(
					constrained_angle,
					_assure_angle_positive(deg2rad(bone_angle_constraints[i].x) + two_bones_back_angle),
					_assure_angle_positive(deg2rad(bone_angle_constraints[i].y) + two_bones_back_angle)
				);
				
				var constrained_vector = Vector2(cos(constrained_angle), sin(constrained_angle));
				
				next_origin = curr_origin + (constrained_vector * direction_length);
				bone_node_transforms[i+1].origin = next_origin;
		# ---- No longer applying joint constraints. ----
		
		var r = next_origin - curr_origin
		var l = bones_in_fabrik_chain_lengths[i] / r.length()
		var bone_final_pos = curr_origin.linear_interpolate(next_origin, l)
		
		# Apply the position and rotation, (potentially with constraints), to the bone node
		bone_node_transforms[i+1] = _transform2d_lookat(bone_node_transforms[i+1], next_origin);
		bone_node_transforms[i+1].origin = bone_final_pos
		
		# Set the joint transform to the node.
		set_bone_transform(i+1, bone_node_transforms[i+1]);


func _calculate_previous_bones_angle(bone_index):
	# Because the transforms are applied to the bone nodes, we can get the global
	# rotation of the bone nodes (which will have parent node rotations applied)
	# and use that to calculate the angle of the bone behind the passed in bone_index.
	if (bone_index-1 >= 0):
		var two_bones_back_angle = bone_nodes[bone_index-1].global_rotation;
		two_bones_back_angle += deg2rad(bones_in_fabrik_chain_angles[bone_index-1]);
		return two_bones_back_angle;
	else:
		return 0;

func _assure_angle_positive(angle_to_convert):
	angle_to_convert = rad2deg(angle_to_convert);
	angle_to_convert = int(round(angle_to_convert * 100)) / 100.0;
	
	if (angle_to_convert < 0):
		while (angle_to_convert < 0):
			angle_to_convert += 360;
		return deg2rad(angle_to_convert);
	if (angle_to_convert > 360):
		while (angle_to_convert < 0):
			angle_to_convert -= 360;
		return deg2rad(angle_to_convert);
	
	return deg2rad(angle_to_convert);


# Works fairly well, in most cases at least..
# TODO: see if I can make this work better.
func _clamp_rotation(angle, min_angle, max_angle):
	if (angle < min_angle or angle > max_angle):
		var min_diff = _angle_difference(angle, min_angle);
		var max_diff = _angle_difference(angle, max_angle);
		
		if (min_diff < max_diff):
			return min_angle;
		else:
			return max_angle;
	else:
		return angle;


# Credit to RNH (comment 15 at URL below)
# http://blog.lexique-du-net.com/index.php?post/Calculate-the-real-difference-between-two-angles-keeping-the-sign
func _angle_difference(angle_1, angle_2):
	var difference = PI - abs(PI - abs(angle_1 - angle_2));
	return difference;


func _transform2d_lookat(p_transform:Transform2D, target:Vector2):
	var local_pos = p_transform.affine_inverse().xform(target);
	var angle_to = (local_pos * p_transform.get_scale()).angle();
	return p_transform.rotated(angle_to);


func set_bone_transform(bone_id, trans):
	bone_nodes[bone_id].global_rotation = trans.get_rotation();
	bone_nodes[bone_id].global_rotation -= deg2rad(bones_in_fabrik_chain_angles[bone_id]);
	
	# NOTE: If the node is a Twisted_Skeleton_Bone, or has the method
	# "_apply_bone_transform_to_skeleton", we need to call it so the
	# iterative approach of FABRIK is preserved and applied correctly.
	if (bone_nodes[bone_id].has_method("_apply_bone_transform_to_skeleton")):
		bone_nodes[bone_id]._apply_bone_transform_to_skeleton();
